from yw_core.yw_core import records2annotations

def check_records2annotations(records, expected_annotations):
    assert records2annotations(records).strip() == expected_annotations.strip()

def test_zeroCell():
    records = []
    expected_annotations = ''
    check_records2annotations(records, expected_annotations)

def test_oneCell_emptyString():
    records = [{'inputs': {}, 'outputs': {}, 'output_candidates': {}}]
    expected_annotations = """
# @BEGIN cell-1
# @END cell-1
"""
    check_records2annotations(records, expected_annotations)

def test_twoCells_emptyString():
    records = [{'inputs': {}, 'outputs': {}, 'output_candidates': {}},
               {'inputs': {}, 'outputs': {}, 'output_candidates': {}}]
    expected_annotations = """
# @BEGIN cell-1
# @END cell-1

# @BEGIN cell-2
# @END cell-2
"""
    check_records2annotations(records, expected_annotations)

def test_oneCell_oneOutputcandidate():
    records = [{'inputs': {}, 'outputs': {}, 'output_candidates': {'x'}}]
    expected_annotations = """
# @BEGIN cell-1
# @END cell-1
"""
    check_records2annotations(records, expected_annotations)

def test_twoCells_cell0zeroInputOneOutput_cell1oneInputzeroOutput():
    records = [{'inputs': {}, 'outputs': {'x'}, 'output_candidates': {}},
               {'inputs': {'x'}, 'outputs': {}, 'output_candidates': {'y'}}]
    expected_annotations = """
# @BEGIN cell-1
# @OUT x
# @END cell-1

# @BEGIN cell-2
# @IN x
# @END cell-2
"""
    check_records2annotations(records, expected_annotations)

def test_twoCells_cell0zeroInputTwoOutputs_cell1twoInputszeroOutput():
    records = [{'inputs': {}, 'outputs': {'x', 'y'}, 'output_candidates': {}},
               {'inputs': {'x', 'y'}, 'outputs': {}, 'output_candidates': {}}]
    expected_annotations = """
# @BEGIN cell-1
# @OUT x
# @OUT y
# @END cell-1

# @BEGIN cell-2
# @IN x
# @IN y
# @END cell-2
"""
    check_records2annotations(records, expected_annotations)

def test_threeCells_cell0zeroInputOneOutput_cell1oneInputOneOutput_cell2oneInputZeroOutput():
    records = [{'inputs': {}, 'outputs': {'x'}, 'output_candidates': {}},
               {'inputs': {'x'}, 'outputs': {'x'}, 'output_candidates': {}},
               {'inputs': {'x'}, 'outputs': {}, 'output_candidates': {}}]
    expected_annotations = """
# @BEGIN cell-1
# @OUT x
# @END cell-1

# @BEGIN cell-2
# @IN x
# @OUT x @AS x-1
# @END cell-2

# @BEGIN cell-3
# @IN x @AS x-1
# @END cell-3
"""
    check_records2annotations(records, expected_annotations)

def test_threeCells_cell0zeroInputOneOutput_cell1empty_cell2oneInputZeroOutput():
    records = [{'inputs': {}, 'outputs': {'x'}, 'output_candidates': {}},
               {'inputs': {}, 'outputs': {}, 'output_candidates': {}},
               {'inputs': {'x'}, 'outputs': {}, 'output_candidates': {}}]
    expected_annotations = """
# @BEGIN cell-1
# @OUT x
# @END cell-1

# @BEGIN cell-2
# @END cell-2

# @BEGIN cell-3
# @IN x
# @END cell-3
"""
    check_records2annotations(records, expected_annotations)



from yw_core.yw_core import extract_records
from textwrap import dedent

EMPTY = set()

def check_extract_records(code_cells, expected_definite_records, expected_possible_records):
    assert extract_records(code_cells=[dedent(code_cell) for code_cell in code_cells], is_upper_estimate=False) == expected_definite_records
    assert extract_records(code_cells=[dedent(code_cell) for code_cell in code_cells], is_upper_estimate=True) == expected_possible_records

def test_zeroCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():
    code_cells = []
    expected_definite_records = []
    expected_possible_records = expected_definite_records
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_oneCellWithEmptyString_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():
    code_cells = ['']
    expected_definite_records = [{'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}]
    expected_possible_records = expected_definite_records
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_varAssigned_andNotUsedInTheSameCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():
    code_cells = ["x = 1"]
    expected_definite_records = [{'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}]
    expected_possible_records = expected_definite_records
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_varAssigned_andUsedInTheSameCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():
    
    code_cells = [
        """
        x = 1
        y = x + 1 
        """
    ]

    expected_definite_records = [{'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_varAssigned_andUsedInAnotherCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        x = 1
        """,""" 
        y = x + 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_varAssigned_andNotUsedInAnotherCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        x = 1
        """,""" 
        y = 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_varAssigned_thenReassignAndUsedInAnotherCell_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        x = 1
        ""","""
        x = 3
        y = x + 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_ifNotInALoop_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        x = 1
        ""","""
        y = 1
        ""","""
        # Union of the if-else branches
        if x > 0:
            tmp1 = 1
            y = y + tmp1
        else:
            tmp2 = 2
            y = y + tmp2
        """
    ]
    
    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'y'}, 'output_candidates': {'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x', 'y'}, 'outputs': EMPTY, 'output_candidates': {'tmp1', 'tmp2', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_ifInALoop_validCellsCorrectDefiniteAnswerWrongPossibleAnswer():
    
    code_cells = [
        """
        x, y = 1, 2
        ""","""
        # Order of executing the if-else branches matters
        # Definite: order with the minimum number of inputs
        # Possible: order with the maximum number of inputs
        for i in range(x):
            if i > 0:
                y = y + a
            else:
                a = 10
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x', 'y'}, 'output_candidates': {'x', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x', 'y'}, 'outputs': EMPTY, 'output_candidates': {'i', 'a', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'x', 'y'}, 'output_candidates': {'x', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x', 'a', 'y'}, 'outputs': EMPTY, 'output_candidates': {'x', 'i', 'a', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_for_validCellsCorrectDefiniteAnswerWrongPossibleAnswer():

    code_cells = [
        """
        x = 5
        ""","""
        # Use x as the argument of range()
        for i in range(x):
            tmp = i + 1
            y = tmp + 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'i', 'tmp', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x', 'i', 'tmp', 'y'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_classInstanceChangedInMethod_validCellsWrongDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        x = [1, 2, 3]
        ""","""
        x.append(1)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}        
    ]

    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_classInstanceNotChangedInMethod_validCellsCorrectDefiniteAnswerWrongPossibleAnswer():

    code_cells = [
        """
        x = [1, 2, 3]
        ""","""
        # The index method does not change x
        # x should NOT be an output candidate
        idx = x.index(1)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'idx'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x', 'idx'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_hiddenModification_mutableObjChangedInPlace_directAssignedMutableObjChangedTogether_validCellsWrongDefiniteAnswerCorrectPossibleAnswer():
    
    code_cells = [
        """
        lst1 = [1, 2, 3]
        lst2 = lst1
        ""","""
        lst1.append(4)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'lst1'}, 'output_candidates': {'lst1', 'lst2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'lst1'}, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'lst1', 'lst2'}, 'output_candidates': {'lst1', 'lst2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst2 = lst1', 'alias_vars': {'lst1', 'lst2'}},
        {'inputs': {'lst1', 'lst2'}, 'outputs': EMPTY, 'output_candidates': {'lst1', 'lst2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst2 = lst1', 'alias_vars': {'lst1', 'lst2'}}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)


def test_hiddenModification_mutableObjChangedInPlace_referredMutableObjsChangedTogether_validCellsWrongDefiniteAnswerCorrectPossibleAnswer():
    
    code_cells = [
        """
        a = set([1, 2, 3])
        b = set([2, 3, 4])
        lst = [a, b]
        ""","""
        for e in lst:
            e.add(1)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'lst'}, 'output_candidates': {'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'lst'}, 'outputs': EMPTY, 'output_candidates': {'e'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'lst', 'a', 'b'}, 'output_candidates': {'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst = [a, b]', 'alias_vars': {'a', 'b', 'lst'}},
        {'inputs': {'lst', 'a', 'b'}, 'outputs': EMPTY, 'output_candidates': {'e', 'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst = [a, b]', 'alias_vars': {'a', 'b', 'lst'}}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_hiddenModification_mutableObjChangedInPlace_referredImmutableObjsDoNotChange_validCellsCorrectDefiniteAnswerWrongPossibleAnswer():
    
    code_cells = [
        """
        a = 1
        b = 2
        lst = [a, b]
        ""","""
        lst[0] = 10
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'lst'}, 'output_candidates': {'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'lst'}, 'outputs': EMPTY, 'output_candidates': {'lst'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'lst', 'a', 'b'}, 'output_candidates': {'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst = [a, b]', 'alias_vars': {'a', 'b', 'lst'}},
        {'inputs': {'lst', 'a', 'b'}, 'outputs': EMPTY, 'output_candidates': {'lst', 'a', 'b'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'lst = [a, b]', 'alias_vars': {'a', 'b', 'lst'}}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_hiddenModification_mutableObjNotChangedInPlace_referredMutableObjsDoNotChange_validCellsWrongDefiniteAnswerWrongPossibleAnswer():
    
    code_cells = [
        """
        import pandas as pd
        df1 = pd.DataFrame({'a': [1, 2, 3]})
        df2 = pd.DataFrame({'b': [2, 3, 4]})
        dfs = [df1, df2]
        ""","""
        for df in dfs:
            df = df + 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': {'dfs'}, 'output_candidates': {'dfs', 'df1', 'df2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'dfs'}, 'outputs': EMPTY, 'output_candidates': {'df'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]

    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': {'dfs', 'df1', 'df2'}, 'output_candidates': {'dfs', 'df1', 'df2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'dfs = [df1, df2]', 'alias_vars': {'df1', 'df2', 'dfs'}},
        {'inputs': {'dfs', 'df1', 'df2'}, 'outputs': EMPTY, 'output_candidates': {'df', 'dfs', 'df1', 'df2'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': 'dfs = [df1, df2]', 'alias_vars': {'df1', 'df2', 'dfs'}}
    ]

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

####################################################################################################
# Assumption 1: inputs and outputs of a function are required to be written at cell level
####################################################################################################

def test_cellLevelVarUsedInNonMutatingFuncBody_notAtCellLevel_validCellsWrongDefiniteAnswerWrongPossibleAnswer_hasWorkaround():

    code_cells = [
        """
        # This non-mutating function uses a variable (i.e., x) defined at cell level
        def add_one(x):
            x = x + 1
            return x
        ""","""
        x = 1
        ""","""
        # Can NOT detect that x is used here
        y = add_one()
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_cellLevelVarUsedInNonMutatingFuncBody_andRedundantlyAtCellLevel_validCellsCorrectDefiniteAnswerWrongPossibleAnswer():

    code_cells = [
        """
        # This non-mutating function uses local variable v
        def add_one(v):
            return v + 1
        ""","""
        x = 1
        ""","""
        # Variable x is used at the cell level
        y = add_one(x)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x', 'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_cellLevelVarUsedInMutatingFuncBody_notAtCellLevel_validCellsWrongDefiniteAnswerWrongPossibleAnswer_hasWorkaround():

    code_cells = [
        """
        # This mutating function uses a variable (i.e., x) defined at cell level
        def add_one():
            x.append(1)
            return x
        ""","""
        x = [1]
        ""","""
        # Can NOT detect that x is used and reassigned here
        y = add_one()
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records

    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_cellLevelVarAssignedInFuncBody_notAtCellLevel_validCellsWrongDefiniteAnswerCorrectPossibleAnswer():

    code_cells = [
        """
        def add_one(v):
            v.append(1)
            return v
        ""","""
        x = [1]
        ""","""
        # Can NOT detect that x is reassigned at this cell
        y = add_one(x)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x', 'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_cellLevelVarUsedAndReassignedInFuncBody_andRedundantlyAtCellLevel_validCellsCorrectDefiniteAnswerCorrectPossibleAnswer():
    
    code_cells = [
        """
        # This mutating function uses local variable v
        def add_one(v):
            v.append(1)
            return v
        ""","""
        x = [1]
        ""","""
        # Varaible x is used and reassigned at cell level
        x = add_one(x)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': {'x'}, 'output_candidates': {'x'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': {'x'}, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

####################################################################################################
# Assumption 2: a variable name can NOT be reused as a function name, and vice versa
####################################################################################################

def test_VarNameReassignedToFunc_ValidCellsWrongDefiniteAnswerWrongPossibleAnswer_hasNoWorkaround():

    code_cells = [
        """
        add_one = 1
        ""","""
        # add_one is not an output candidate as it is reassigned to a function
        add_one = add_one + 1
        # Reassign the variable name to a function
        def add_one(v):
            return v + 1
        y = add_one(5)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'add_one'}, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_funcNameReassignedToVar_ValidCellsWrongDefiniteAnswerWrongPossibleAnswer_hasNoWorkaround():

    code_cells = [
        """
        def add_one(v):
            return v + 1
        ""","""
        x = add_one(5)
        # Reassign the function name to a variable
        add_one = x + 1
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'x'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)

def test_funcNameReassignedToVarAndReused_InvalidCellsAnyAnswer():

    code_cells = [
        """
        def add_one(v):
            return v + 1
        ""","""
        # Reassign the function name to a variable
        add_one = 5
        ""","""
        # The function cannot be reused as it has been reassigned to a variable.
        y = add_one(1)
        """
    ]

    expected_definite_records = [
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': {'add_one'}, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': EMPTY, 'refers_code': EMPTY, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY},
        {'inputs': EMPTY, 'outputs': EMPTY, 'output_candidates': {'y'}, 'refers_code': {'add_one'}, 'defines_code': EMPTY, 'alias_stmt': None, 'alias_vars': EMPTY}
    ]
    expected_possible_records = expected_definite_records
    
    check_extract_records(code_cells, expected_definite_records, expected_possible_records)



import nbformat, ast, copy, re, sys
from pathlib import Path

def clean_code(cell_code):
    # Remove syntaxes for line magics (%time, %timeit) along with any options (-r2, -n10, etc.)
    cleaned_code = re.sub(r"^\s*%time(it)?\s+(-\S+\s+)*", "", cell_code.strip(), flags=re.MULTILINE)
    # Remove syntaxes for cell magics (%%time, %%timeit)
    cleaned_code = re.sub(r"^\s*%%time(it)?\s+(-\S+\s+)*", "", cleaned_code, flags=re.MULTILINE)

    # Remove cells that start with %%
    if cleaned_code.startswith("%%"):
        cleaned_code = ""
    else: # Remove lines starts with % or !
        cleaned_code = re.sub(r"^\s*[%!].*\n?", "", cleaned_code, flags=re.MULTILINE)
    
    return cleaned_code

def extract_code_cells(notebook_path):
    """
    Extract code cells from a given Jupyter Notebook.
    :param notebook_path: String. Path of a Jupyter Notebook.
    :return code_cells: a list of code cells.
    """
    with open(notebook_path, 'r') as f:
        notebook = nbformat.read(f, as_version=4)
    code_cells = [clean_code(cell['source']) for cell in notebook.cells if cell.cell_type == 'code']
    return code_cells

class MinIOVisitor(ast.NodeVisitor):
    def __init__(self, alias_mapping={}, modules=set()):
        self.inputs = set()
        self.outputs = set()
        self.refers_code = set()
        self.defines_code = set()
        self.alias_mapping = alias_mapping
        self.modules = modules
        self.loopflag = False
    
    def visit_Import(self, node):
        for alias in node.names:
            if alias.asname:
                self.modules.add(alias.asname)
            else:
                self.modules.add(alias.name)
    
    def visit_ImportFrom(self, node):
        self.visit_Import(node)
    
    def _visit_list_of_nodes(self, nodes):
        for item in nodes:
            prev_inputs, prev_outputs = self.inputs.copy(), self.outputs.copy()
            self.visit(item)
            # Exclude previous outputs from current inputs
            curr_inputs = self.inputs - prev_inputs
            self.inputs = (curr_inputs - prev_outputs) | prev_inputs
    
    def visit_Try(self, node):
        # Union of the try-except-finally blocks
        body_inputs, body_outputs = set(), set()
        for item in node.handlers:
            self._visit_list_of_nodes(node.body + item.body + node.finalbody)
            body_inputs.update(self.inputs)
            body_outputs.update(self.outputs)
            self.inputs, self.outputs = set(), set()
        self._visit_list_of_nodes(node.body + node.orelse + node.finalbody)
        body_inputs.update(self.inputs)
        body_outputs.update(self.outputs)
    
    def visit_TryStar(self, node):
        self.visit_Try(node)

    def visit_If(self, node):
        prev_inputs, prev_outputs = self.inputs.copy(), self.outputs.copy()
        # In a loop
        if self.loopflag:
            # Order 1
            self._visit_list_of_nodes([node.test] + node.body + node.orelse)
            order1_inputs, order1_outputs = self.inputs.copy(), self.outputs.copy()
            self.inputs, self.outputs = prev_inputs.copy(), prev_outputs.copy()
            # Order 2
            self._visit_list_of_nodes([node.test] + node.orelse + node.body)
            # Keep the intersection of these two orders
            self.inputs = order1_inputs & self.inputs
            self.outputs = order1_outputs & self.outputs
        else:
            # Union of the if-else branches
            self._visit_list_of_nodes([node.test] + node.body)
            union_inputs, union_outputs = self.inputs.copy(), self.outputs.copy()
            self.inputs, self.outputs = prev_inputs.copy(), prev_outputs.copy()
            self._visit_list_of_nodes([node.test] + node.orelse)
            self.inputs.update(union_inputs)
            self.outputs.update(union_outputs)
    
    def visit_For(self, node):
        self.loopflag = True
        self._visit_list_of_nodes([node.iter, node.target] + node.body + node.orelse)
    
    def visit_AsyncFor(self, node):
        self.loopflag = True
        self._visit_list_of_nodes([node.iter, node.target] + node.body + node.orelse)
    
    def visit_While(self, node):
        self.loopflag = True
        self._visit_list_of_nodes([node.test] + node.body + node.orelse)
    
    def visit_comprehension(self, node):
        self._visit_list_of_nodes([node.iter, node.target] + node.ifs)
    
    def visit_ListComp(self, node):
        self._visit_list_of_nodes(node.generators + [node.elt])
    
    def visit_SetComp(self, node):
        self._visit_list_of_nodes(node.generators + [node.elt])
    
    def visit_DictComp(self, node):
        self._visit_list_of_nodes(node.generators + [node.key, node.value])
    
    def visit_GeneratorExp(self, node):
        self._visit_list_of_nodes(node.generators + [node.elt])

    def visit_With(self, node):
        self._visit_list_of_nodes(node.items + node.body)
    
    def visit_AugAssign(self, node):
        if isinstance(node.target, ast.Name):
            """
            Add the variable as an input as well
            Example: print(ast.dump(ast.parse('a += 1'), indent=4))
            # a is both an input and an output
            Module(
                body=[
                    AugAssign(
                        target=Name(id='a', ctx=Store()),
                        op=Add(),
                        value=Constant(value=1))],
                type_ignores=[])
            """
            self.inputs.add(node.target.id)
        self.generic_visit(node)

    def _visit_Name_add2inorout(self, node):
        if isinstance(node.ctx, ast.Load):
            self.inputs.add(node.id)
        elif isinstance(node.ctx, (ast.Store, ast.Del)):
            self.outputs.add(node.id)
    
    def _visit_Name_add2in(self, node):
        self.inputs.add(node.id)
        self._visit_Name_add2inorout(node)
        
    def _visit_Name_add2out(self, node):
        self.outputs.add(node.id)
        self._visit_Name_add2inorout(node)
    
    def _visit_Name_add2inandout(self, node):
        self.inputs.add(node.id)
        self.outputs.add(node.id)

    def visit_Name(self, node):
        """
        Reference: https://docs.python.org/3/library/ast.html#ast.Name
        Example: print(ast.dump(ast.parse('b = a + 1'), indent=4))
        Module(
            body=[
                Assign(
                    targets=[
                        Name(id='b', ctx=Store())],
                    value=BinOp(
                        left=Name(id='a', ctx=Load()),
                        op=Add(),
                        right=Constant(value=1)))],
            type_ignores=[])
        """
        self._visit_Name_add2inorout(node)
    
    def visit_AsyncFunctionDef(self, node):
        self.defines_code.add(node.name)

    def visit_FunctionDef(self, node):
        self.defines_code.add(node.name)
    
    def visit_ClassDef(self, node):
        self.defines_code.add(node.name)
    
    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            self.refers_code.add(node.func.id)
        self.generic_visit(node)
    
    def visit_Lambda(self, node):
        self.generic_visit(node.body)
        for arg in node.args.args:
            if arg.arg in self.inputs:
                self.inputs.remove(arg.arg)
            if arg.arg in self.outputs:
                self.outputs.remove(arg.arg)

    def _reset_name_for_store_and_del(self, prev_visit_name_func):
        if prev_visit_name_func != self._visit_Name_add2inandout:
            if self.visit_Name == self._visit_Name_add2in:
                self.visit_Name = self._visit_Name_add2inandout
            else:
                self.visit_Name = self._visit_Name_add2out

    def _reset_name_for_load(self, prev_visit_name_func):
        if prev_visit_name_func != self._visit_Name_add2inandout:
            if self.visit_Name == self._visit_Name_add2out:
                self.visit_Name = self._visit_Name_add2inandout
            else:
                self.visit_Name = self._visit_Name_add2in

    def visit_Attribute(self, node):
        prev_visit_name_func = self.visit_Name
        if isinstance(node.ctx, (ast.Store, ast.Del)):
            self._reset_name_for_store_and_del(prev_visit_name_func)
            self.generic_visit(node)
            self.visit_Name = prev_visit_name_func
        if isinstance(node.ctx, ast.Load):
            self._reset_name_for_load(prev_visit_name_func)
            self.generic_visit(node)
            self.visit_Name = prev_visit_name_func

    def visit_Subscript(self, node):
        if isinstance(node.ctx, (ast.Store, ast.Del)):
            self.visit(node.slice)
            prev_visit_name_func = self.visit_Name
            self._reset_name_for_store_and_del(prev_visit_name_func)
            self.visit(node.value)
            self.visit_Name = prev_visit_name_func
        if isinstance(node.ctx, ast.Load):
            prev_visit_name_func = self.visit_Name
            self._reset_name_for_load(prev_visit_name_func)
            self.generic_visit(node)
            self.visit_Name = prev_visit_name_func

    def visit(self, node):
        super().visit(node)
        return self.inputs - self.refers_code, self.outputs - self.refers_code, self.refers_code, self.defines_code, self.alias_mapping, self.modules

class MaxIOVisitor(MinIOVisitor):

    def _reset_name_before(self, prev_visit_name_func):
        self.visit_Attribute = self.generic_visit
        self.visit_Subscript = self.generic_visit
        super()._reset_name_for_store_and_del(prev_visit_name_func)
    
    def _reset_name_after(self, prev_visit_name_func):
        self.visit_Name = prev_visit_name_func
        self.visit_Attribute = super().visit_Attribute
        self.visit_Subscript = super().visit_Subscript

    def visit_Assign(self, node):
        for target in node.targets:
            if isinstance(target, ast.Name):
                items = set()
                # Direct assignment
                if isinstance(node.value, ast.Name):
                    items.add(node.value.id)
                # List or Tuple
                elif isinstance(node.value, (ast.List, ast.Tuple)):
                    for item in node.value.elts:
                        if isinstance(item, ast.Name):
                            items.add(item.id)
                # Dict
                elif isinstance(node.value, ast.Dict):
                    for item in node.value.values:
                        if isinstance(item, ast.Name):
                            items.add(item.id)
                if target.id in self.alias_mapping and self.alias_mapping[target.id]:
                    # Remove previous references: items -> target.id
                    for item in self.alias_mapping[target.id]:
                        self.alias_mapping[item].remove(target.id)
                # Overwrite the target.id mapping: target.id -> items
                self.alias_mapping[target.id] = items.copy()
                # Update the mapping for the items: item -> target.id
                for item in items:
                    self.alias_mapping[item] = self.alias_mapping.get(item, set()) | {target.id}
                # Remove empty mappings in the alias_mapping
                self.alias_mapping = {k: v for k, v in self.alias_mapping.items() if v}
        self.generic_visit(node)

    # Rule: if a variable is used as an iterator in a for, a while loop, or a comprehension, 
    # it is regarded as both input and output
    def _visit_list_of_nodes(self, nodes):
        for idx, item in enumerate(nodes):
            prev_inputs, prev_outputs = self.inputs.copy(), self.outputs.copy()
            if idx == 0:
                prev_visit_name_func = self.visit_Name
                # Handle the iter node (for loop) or the test node (while loop) or the iter node (comprehension)
                self._reset_name_before(prev_visit_name_func)
                self.visit(item)
                self._reset_name_after(prev_visit_name_func)
            else:
                self.visit(item)
            # Exclude outputs from previous iterations
            curr_inputs = self.inputs - prev_inputs
            self.inputs = (curr_inputs - prev_outputs) | prev_inputs
    
    def _visit_list_of_nodes_in_if(self, nodes):
        # visit the super calss method
        super()._visit_list_of_nodes(nodes)
    
    def visit_If(self, node):
        prev_inputs, prev_outputs = self.inputs.copy(), self.outputs.copy()
        # In a loop
        if self.loopflag:
            # Order 1
            self._visit_list_of_nodes_in_if([node.test] + node.body + node.orelse)
            order1_inputs, order1_outputs = self.inputs.copy(), self.outputs.copy()
            self.inputs, self.outputs = prev_inputs.copy(), prev_outputs.copy()
            # Order 2
            self._visit_list_of_nodes_in_if([node.test] + node.orelse + node.body)
            # Keep the union of these two orders
            self.inputs = order1_inputs | self.inputs
            self.outputs = order1_outputs | self.outputs
        else:
            # Union of the if-else branches
            self._visit_list_of_nodes_in_if([node.test] + node.body)
            union_inputs, union_outputs = self.inputs.copy(), self.outputs.copy()
            self.inputs, self.outputs = prev_inputs.copy(), prev_outputs.copy()
            self._visit_list_of_nodes_in_if([node.test] + node.orelse)
            self.inputs.update(union_inputs)
            self.outputs.update(union_outputs)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            self.refers_code.add(node.func.id)
        elif isinstance(node.func, ast.Attribute):
            while isinstance(node.func.value, ast.Attribute):
                node.func = node.func.value
        prev_visit_name_func = self.visit_Name
        self._reset_name_before(prev_visit_name_func)
        self.generic_visit(node)
        self._reset_name_after(prev_visit_name_func)

def _add_refers(data, alias_mapping):
    for node in copy.deepcopy(data):
        if node in alias_mapping:
            data.update(alias_mapping[node])

def parse_per_statement(node, alias_mapping, modules, is_upper_estimate):
    Visitor = MaxIOVisitor(alias_mapping, modules) if is_upper_estimate else MinIOVisitor(alias_mapping, modules)
    return Visitor.visit(node)

def capture_variables(code_cell, alias_mapping={}, modules=set(), is_upper_estimate=False, parse_per_statement=parse_per_statement):
    cell_inputs, cell_outputs, cell_refers_code, cell_defines_code = set(), set(), set(), set()
    # Parse the code cell
    tree = ast.parse(code_cell, mode='exec', type_comments=False)
    excluded_inputs = set()
    prev_alias_mapping = copy.deepcopy(alias_mapping)
    alias_stmt = None
    for node in tree.body:
        # For each statement in a code cell
        inputs, outputs, refers_code, defines_code, alias_mapping, modules = parse_per_statement(node, alias_mapping, modules, is_upper_estimate)
        curr_inputs = inputs - excluded_inputs
        if prev_alias_mapping == alias_mapping:
            _add_refers(curr_inputs, alias_mapping)
            _add_refers(outputs, alias_mapping)
        else:
            alias_stmt = alias_stmt + '\n' + ast.unparse(node) if alias_stmt else ast.unparse(node)
        prev_alias_mapping = copy.deepcopy(alias_mapping)
        cell_inputs.update(curr_inputs)
        cell_outputs.update(outputs)
        cell_refers_code.update(refers_code)
        cell_defines_code.update(defines_code)
        excluded_inputs.update(outputs)
    return cell_inputs, cell_outputs, cell_refers_code, cell_defines_code, alias_mapping, alias_stmt, modules

def extract_records(code_cells, is_upper_estimate=False, parse_per_statement=parse_per_statement):
    """
    Extract inputs, outputs, and output_candidates from a list of code cells.
    :param code_cells: a list of code cells.
    :return records: a list of records (dict). Records are in reversed order for each code cell, containing inputs, outputs, and output_candidates.
    """
    # Consider alias_mapping in order
    # and exclude used funcs and classes (refers_code) that are not defined in the previous code cells
    alias_mapping, modules, records = {}, set(), []
    defined_code = set()
    prev_alias_stmt = None
    for idx, code_cell in enumerate(code_cells):
        inputs, output_candidates, refers_code, defines_code, alias_mapping, alias_stmt, modules = capture_variables(code_cell, alias_mapping=alias_mapping, modules=modules, is_upper_estimate=is_upper_estimate, parse_per_statement=parse_per_statement)
        defined_code = defined_code | defines_code
        prev_alias_stmt = alias_stmt if alias_stmt else prev_alias_stmt
        alias_vars = set(alias_mapping.keys()).union(*alias_mapping.values())
        records.append({'inputs': inputs - defined_code, 'output_candidates': output_candidates - defined_code, 'refers_code': (inputs | refers_code) & defined_code, 'defines_code': defines_code, 'alias_stmt': prev_alias_stmt, 'alias_vars': alias_vars})

    # Only keep variables that are used in the subsequent code cells as outputs
    required_variables = set()
    for idx in range(len(records)-1, -1, -1): # reverse order
        inputs, output_candidates = records[idx]['inputs'], records[idx]['output_candidates']
        outputs = output_candidates & required_variables
        required_variables = (required_variables - outputs) | inputs
        records[idx]['outputs'] = outputs

    # Exclude imported modules (e.g., pd) and built-in functions/variables (e.g.,len and __name__) from inputs, outputs, and output_candidates
    with open(Path(__file__).parent / "builtin_funcs_and_vars.txt", "r") as fin:
        builtin_funcs_and_vars = set(fin.read().splitlines()[1:])
    modules_and_builtins = modules | builtin_funcs_and_vars
    existing_outputs = set()
    for idx, record in enumerate(records):
        # If a variable is not defined in the previous code cells
        # and is one of the imported modules or built-in functions/variables
        excluded_vars = (record['inputs'] - existing_outputs) & modules_and_builtins
        inputs = record['inputs'] - excluded_vars
        outputs = record['outputs'] - excluded_vars
        output_candidates = record['output_candidates'] - excluded_vars
        records[idx]['inputs'], records[idx]['outputs'], records[idx]['output_candidates'] = inputs, outputs, output_candidates
        existing_outputs.update(outputs)
    return records

def map_records(records):
    """
    Map records to avoid duplicate names in inputs and outputs.
    :param records: a list of records (dict). Records are in the same order as code cells, containing inputs, outputs, and output_candidates.
    :return records: a list of records (dict). Records are in the same order as code cells, containing inputs and outputs.
    """
    # Rename generated output with -{v} suffix, where v is 1 OR 2 OR 3, ...
    name_mappings = {}
    for record in records:
        mapped_inputs, mapped_outputs = set(), set()
        for input in record['inputs']:
            if (input in name_mappings and name_mappings[input] == 0) or (input not in name_mappings):
                name_mappings[input] = 0
                mapped_inputs.add(input)
            else:
                mapped_inputs.add(input + '-' + str(name_mappings[input]))
        for output in record['outputs']:
            if output in name_mappings:
                name_mappings[output] = name_mappings[output] + 1
                mapped_outputs.add(output + '-' + str(name_mappings[output]))
            else:
                name_mappings[output] = 0
                mapped_outputs.add(output)
        record['inputs'], record['outputs'] = mapped_inputs, mapped_outputs
    return records

def get_input_or_output_annotations(vars, is_input):
    """
    Generate YesWorkflow Syntax from a list of variables.
    :param vars: a set of variables.
    :param is_input: Boolean. True if vars are inputs, False if vars are outputs.
    :return annotations: String. YesWorkflow annotations.
    """
    annotations = ''
    for var in sorted(vars):
        if is_input:
            if '-' in var:
                annotations = annotations + "# @IN {var} @AS {alias}\n".format(var=var.split('-')[0], alias=var)
            else:
                annotations = annotations + "# @IN {var}\n".format(var=var)
        else:
            if '-' in var:
                annotations = annotations + "# @OUT {var} @AS {alias}\n".format(var=var.split('-')[0], alias=var)
            else:
                annotations = annotations + "# @OUT {var}\n".format(var=var)
    return annotations

def records2annotations(records):
    """
    Generate YesWorkflow Syntax from a list of records about varaibles.
    :param records: a list of records (dict). Records are in the same order as code cells, containing inputs, outputs, and output_candidates.
    :return annotations: String. YesWorkflow annotations.
    """
    annotations = ''
    mapped_records = map_records(records)
    max_idx = len(records) - 1
    for idx, record in enumerate(mapped_records):
        inputs, outputs = record['inputs'], record['outputs']
        annotations = annotations + "# @BEGIN cell-{idx}\n".format(idx=idx+1) + \
        get_input_or_output_annotations(inputs, True) + \
        get_input_or_output_annotations(outputs, False) + \
        "# @END cell-{idx}".format(idx=idx+1)
        if idx < max_idx:
            annotations = annotations + "\n\n"
    return annotations

def yw_core(code_cells):
    records = extract_records(code_cells, is_upper_estimate=False)
    annotations = records2annotations(copy.deepcopy(records))
    return annotations, records



from yw_core.yw_core import extract_code_cells, extract_records, records2annotations
import click, sys, os

@click.command()
@click.option('--filepath', '-f', required=True, type=str, help='Path of a Python notebook to extract YesWorkflow annotations')
@click.option('--upper', '-u', is_flag=True, show_default=True, default=False, help='Extract upper estimate of cell I/O sets for YesWorkflow annotations')
def main(filepath, upper):
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"File '{filepath}' not found")
    code_cells = extract_code_cells(filepath)
    records = extract_records(code_cells, is_upper_estimate=upper)
    annotations = records2annotations(records)
    print(annotations)
